/*
 * Copyright (c) 2002, mulix@actcom.co.il
 *
 *       discd $Id: discd.c,v 1.2 2002/02/18 21:11:01 mulix Exp $
 *
 * Permission to use, copy and distribute this software is hereby granted
 * under the terms of version 2 or any later version of the GNU General Public
 * License, as published by the Free Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED IN ITS "AS IS" CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */

/*
 * Author(s): mulix
 */

/* code based in parts upon code by the mosix team (setpe), */
/* moshe bar (multicast), the syscalltrack project and fortuned */
/* (daemonize), all GPL'd. */ 

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include "util.h"

#define OPEN_MOSIX_DIR "/proc/mosix/admin"
#define OPEN_MOSIX_CONFIG_FILE OPEN_MOSIX_DIR "/config"
#define NODE_DISC_COOKIE 0xc0debabe

#define MSG_I_AM_HERE "hello"
#define MSG_WELCOME "welcome"

/* these should be specified outside somehow */ 
#define MCAST_PORT 12345 
#define MCAST_GROUP "224.3.4.5"

static void 
check_mosix(void)
{
	struct stat sbuf; 

	/* see if mosix proc dir exists */ 
	if (stat(OPEN_MOSIX_DIR, &sbuf))
		fatal("Error: This is NOT mosix! (%s doesnt exist)", 
		      OPEN_MOSIX_DIR); 
    
	if (!S_ISDIR(sbuf.st_mode))
		fatal("Error: mosix proc dir (%s) exists but "
		      "is not a directory?!", OPEN_MOSIX_DIR); 

	if (access(OPEN_MOSIX_CONFIG_FILE, W_OK) == -1)
		fatal("Error: no write access to '%s'", 
		      OPEN_MOSIX_CONFIG_FILE); 
}

static void 
announce_i_am_here(void)
{
	struct sockaddr_in addr;
	int fd, sz; 
	int num_tries = 3; 
	char message[32]; 
    
	if ((sz = snprintf(message, sizeof(message), "%x%s", 
			   NODE_DISC_COOKIE, MSG_I_AM_HERE)) == -1) {
		fatal("welcome message was truncated. recompile me"); 
	}

	if ((fd = socket(AF_INET,SOCK_DGRAM,0)) < 0) {
		fatal("socket: %s", strerror(errno)); 
	}

	memset(&addr,0,sizeof(addr));
	addr.sin_family=AF_INET;
	addr.sin_addr.s_addr=inet_addr(MCAST_GROUP);
	addr.sin_port=htons(MCAST_PORT);

	while (1) {
		if (sendto(fd, message, sz, 0,(struct sockaddr *) &addr,
			   sizeof(addr)) < 0) {
			if (!--num_tries) {
				fatal("sendto: %s", strerror(errno)); 
			}
			sleep(1);
			continue; 
		}
		/* whee, send was succesfull, break */ 
		break; 
	}

	close(fd); 
}

/* the functio used to turn us into a daemon */
static void 
daemonize(void)
{
	pid_t pid;
    
	/* fork for the first time. */ 
	if ((pid = fork()) != 0){
		if (pid == -1){
			fatal("failed to fork into background! error '%s'", 
			      strerror(errno));
		}
		/* parent, exit */
		exit(0);
	}

	/* we're now in the first child */
	/* setsid() will create a new session for us, since 
	   because of the earlier fork() we're not a process group leader.
	   this will assure that we have no controlling terminal */
	setsid();

	/* ignore SIGHUP signals */
	signal(SIGHUP, SIG_IGN);
    
	/* fork again */
	/* this guarantees that we will never have a controlling terminal, */
	/* since we never want be the session group leader anymore         */
	if ((pid = fork()) != 0){
		if (pid == -1){
			fatal("failed to fork into background! error '%s'", 
			      strerror(errno));
		}
		/* parent, exit */
		exit(0);
	}

	/* change our working directory to /, to make sure */
	/* umounting the file system where we started is possible */
	chdir("/");
     
	/* set out umask to 0, to make sure we fully control our */
	/* file creating bits */
	umask(0);

	/* close all (possibly) open file desriptors */
	/* TODO: calling getdtablesize() here loops endlessly. why? */
#if 0
	for (i = 0; i < MAXFD ; i++)
		close(i);
#endif 
	close(fileno(stdin)); 
	close(fileno(stdout)); 
	close(fileno(stderr)); 
    
	log("succesfully daemonized");
}

static void 
update_mosix_map(struct sockaddr_in* addr, unsigned int addrlen)
{
	/* TODO */
	/* we get the the src addr, which we need to extract the ip from */
	/* are we guaranteed it's the ip of the sending host, and not 
	   that of the multicast address? */
}

static void 
handle_incoming_message(char* msgbuf, size_t bufsz, 
			struct sockaddr_in* addr, 
			unsigned int addrlen)
{
	/* what should be the structure of an incoming message? */
	/* A: first the cookie, and then "welcome" */
	/* we get the source address from addr */
	unsigned int cookie; 

	if (bufsz < sizeof(cookie) + sizeof(MSG_WELCOME)) {
		warn("got an illegal packet!"); /* TODO: from whom, with what contents? */ 
		return;
	}

	memcpy(&cookie, msgbuf, sizeof(cookie)); 

	if (cookie != NODE_DISC_COOKIE) {
		warn("cookie: expected %x, got %x", NODE_DISC_COOKIE, cookie); 
		return;
	}

	if (strncmp(msgbuf + sizeof(cookie), 
		    MSG_WELCOME, 
		    bufsz - sizeof(cookie))) {
		warn("packet with right cookie but wrong data!"); 
		return;
	}

	/* WOOHOO, success! */
	update_mosix_map(addr, addrlen); 
}

static void 
listen_for_announcements(void)
{
#define MSGBUFSIZE 256
    
	struct sockaddr_in addr;
	int fd, nbytes,addrlen;
	char msgbuf[MSGBUFSIZE+1];
	struct ip_mreq mreq;

	/* create what looks like an ordinary UDP socket */
	if ((fd=socket(AF_INET,SOCK_DGRAM,0)) < 0) {
		fatal("socket: %s", strerror(errno));
	}

	/* set up destination address */
	memset(&addr,0,sizeof(addr));
	addr.sin_family=AF_INET;
	addr.sin_addr.s_addr=htonl(INADDR_ANY); 
	addr.sin_port=htons(MCAST_PORT);
    
	if (bind(fd,(struct sockaddr *) &addr,sizeof(addr)) < 0) {
		fatal("bind: %s", strerror(errno)); 
	}

	/* use setsockopt() to request that the kernel join a multicast group */
	memset(&mreq, 0, sizeof(mreq)); 
	mreq.imr_multiaddr.s_addr=inet_addr(MCAST_GROUP);
	mreq.imr_interface.s_addr=htonl(INADDR_ANY);
	if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0) {
		fatal("setsockopt: %s", strerror(errno));
	}

	addrlen = sizeof(addr);
	while (1) {
		if ((nbytes=recvfrom(fd, msgbuf, MSGBUFSIZE, 0,
				     (struct sockaddr *) &addr,&addrlen)) < 0) {
			if (errno == EINTR || errno == EAGAIN)
				continue; 
	    
			fatal("recvfrom: %s", strerror(errno)); 
		}
		msgbuf[nbytes] = '\0'; /* null terminate the buffer */ 

		/* we assume that recvfrom() will return the entire datagram */
		handle_incoming_message(msgbuf, nbytes, &addr, addrlen); 
	}
}

int 
main(int argc, char** argv)
{
	/* check that we're running on a mosix system */ 
	check_mosix();

	/* tell the other nodes that we're here */ 
	announce_i_am_here(); 

	/* become a daemon */ 
	daemonize(); 
    
	/* wait for messages */
	listen_for_announcements(); 
    
	/* never reached */ 
	return 0; 
}
